// src/types.ts
var WebCodecsEncoderError = class _WebCodecsEncoderError extends Error {
  constructor(type, message, cause) {
    super(message);
    this.type = type;
    this.cause = cause;
    this.name = "WebCodecsEncoderError";
    Object.setPrototypeOf(this, _WebCodecsEncoderError.prototype);
  }
};

// src/logger.ts
var logger = {
  log: (...args) => {
    console.log(...args);
  },
  warn: (...args) => {
    console.warn(...args);
  },
  error: (...args) => {
    console.error(...args);
  }
};
var logger_default = logger;

// src/encoder.ts
function validateEncoderConfig(config) {
  if (config.width <= 0 || config.width > 7680) {
    throw new WebCodecsEncoderError(
      "validation-error" /* ValidationError */,
      `Invalid width: ${config.width}. Must be between 1 and 7680 pixels.`
    );
  }
  if (config.height <= 0 || config.height > 4320) {
    throw new WebCodecsEncoderError(
      "validation-error" /* ValidationError */,
      `Invalid height: ${config.height}. Must be between 1 and 4320 pixels.`
    );
  }
  if (config.frameRate <= 0 || config.frameRate > 120) {
    throw new WebCodecsEncoderError(
      "validation-error" /* ValidationError */,
      `Invalid frameRate: ${config.frameRate}. Must be between 0.1 and 120 fps.`
    );
  }
  if (config.videoBitrate < 1e5 || config.videoBitrate > 1e8) {
    throw new WebCodecsEncoderError(
      "validation-error" /* ValidationError */,
      `Invalid videoBitrate: ${config.videoBitrate}. Must be between 100kbps and 100Mbps.`
    );
  }
  if (config.audioBitrate > 0) {
    if (config.channels <= 0 || config.channels > 8) {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Invalid channels: ${config.channels}. Must be between 1 and 8.`
      );
    }
    if (config.sampleRate < 8e3 || config.sampleRate > 192e3) {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Invalid sampleRate: ${config.sampleRate}. Must be between 8kHz and 192kHz.`
      );
    }
    if (config.audioBitrate < 32e3 || config.audioBitrate > 32e4) {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Invalid audioBitrate: ${config.audioBitrate}. Must be between 32kbps and 320kbps.`
      );
    }
  }
  const container = config.container ?? "mp4";
  const videoCodec = config.codec?.video ?? (container === "webm" ? "vp9" : "avc");
  const audioCodec = config.codec?.audio ?? (container === "webm" ? "opus" : "aac");
  if (container === "webm") {
    if (!["vp8", "vp9", "av1"].includes(videoCodec)) {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Video codec '${videoCodec}' is not compatible with WebM container. Use vp8, vp9, or av1.`
      );
    }
    if (audioCodec === "aac") {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Audio codec 'aac' is not compatible with WebM container. Use opus.`
      );
    }
  } else if (container === "mp4") {
    if (!["avc", "hevc", "av1"].includes(videoCodec)) {
      throw new WebCodecsEncoderError(
        "validation-error" /* ValidationError */,
        `Video codec '${videoCodec}' is not compatible with MP4 container. Use avc, hevc, or av1.`
      );
    }
  }
}
var WebCodecsEncoder = class _WebCodecsEncoder {
  constructor(config) {
    this.worker = null;
    this.processedFramesInternal = 0;
    this.submittedFramesInternal = 0;
    this.droppedFramesInternal = 0;
    this.videoQueueSizeInternal = 0;
    this.audioQueueSizeInternal = 0;
    this.actualVideoCodec = null;
    this.actualAudioCodec = null;
    // 状態管理
    this.currentState = "idle" /* Idle */;
    this.currentStage = "initializing" /* Initializing */;
    // 詳細プログレス追跡
    this.processingStartTime = 0;
    this.lastProgressTime = 0;
    this.lastProcessedFrames = 0;
    this.processedDataSize = 0;
    this.processingFpsHistory = [];
    this.maxFpsHistoryLength = 10;
    // Callbacks for asynchronous operations
    this.onInitialized = null;
    this.onInitializeError = null;
    this.onFinalizedPromise = null;
    this.onProgressCallback = null;
    this.onDetailedProgressCallback = null;
    this.onErrorCallback = null;
    this.onDataCallback = null;
    // For real-time data
    this.isCancelled = false;
    this.nextVideoTimestamp = 0;
    this.nextAudioTimestamp = 0;
    this.audioContext = null;
    this.audioWorkletNode = null;
    this.cancelTimeoutId = null;
    if (config) {
      validateEncoderConfig(config);
      this.config = {
        // Default values first
        container: "mp4",
        latencyMode: "quality",
        dropFrames: false,
        maxQueueDepth: Infinity,
        ...config,
        // User-provided config overrides defaults
        codec: {
          // Ensure codec object exists and has defaults
          video: config.codec?.video ?? "avc",
          // Use 'avc' as per type
          audio: config.codec?.audio ?? "aac"
          // Use 'avc' as per type
        }
      };
    } else {
      this.config = {};
    }
  }
  static isSupported() {
    return typeof VideoEncoder !== "undefined" && typeof AudioEncoder !== "undefined" && typeof Worker !== "undefined";
  }
  /**
   * メソッドチェーン用のファクトリーメソッド
   */
  static create() {
    return new _WebCodecsEncoder();
  }
  /**
   * 現在の状態を取得
   */
  getState() {
    return this.currentState;
  }
  /**
   * 現在の処理ステージを取得
   */
  getCurrentStage() {
    return this.currentStage;
  }
  /**
   * 指定された状態でのAPI呼び出しが許可されているかチェック
   */
  checkStateForOperation(operation, allowedStates) {
    if (!allowedStates.includes(this.currentState)) {
      throw new WebCodecsEncoderError(
        "internal-error" /* InternalError */,
        `Cannot ${operation} in state '${this.currentState}'. Allowed states: ${allowedStates.join(", ")}`
      );
    }
  }
  /**
   * 状態を変更
   */
  setState(newState) {
    logger_default.log(
      `WebCodecsEncoder: State transition: ${this.currentState} -> ${newState}`
    );
    this.currentState = newState;
  }
  /**
   * 処理ステージを変更
   */
  setStage(newStage) {
    logger_default.log(
      `WebCodecsEncoder: Stage transition: ${this.currentStage} -> ${newStage}`
    );
    this.currentStage = newStage;
  }
  // === メソッドチェーン用API ===
  /**
   * エンコーダーの設定を行う（メソッドチェーン用）
   */
  configure(config) {
    this.checkStateForOperation("configure", ["idle" /* Idle */]);
    validateEncoderConfig(config);
    this.config = {
      // Default values first
      container: "mp4",
      latencyMode: "quality",
      dropFrames: false,
      maxQueueDepth: Infinity,
      ...config,
      // User-provided config overrides defaults
      codec: {
        // Ensure codec object exists and has defaults
        video: config.codec?.video ?? "avc",
        // Use 'avc' as per type
        audio: config.codec?.audio ?? "aac"
        // Use 'aac' as per type
      }
    };
    return this;
  }
  /**
   * プログレスコールバックを設定（メソッドチェーン用）
   */
  onProgress(callback) {
    this.onProgressCallback = callback;
    return this;
  }
  /**
   * 詳細プログレスコールバックを設定（メソッドチェーン用）
   */
  onDetailedProgress(callback) {
    this.onDetailedProgressCallback = callback;
    return this;
  }
  /**
   * エラーコールバックを設定（メソッドチェーン用）
   */
  onError(callback) {
    this.onErrorCallback = callback;
    return this;
  }
  /**
   * リアルタイムデータコールバックを設定（メソッドチェーン用）
   */
  onData(callback) {
    this.onDataCallback = callback;
    return this;
  }
  /**
   * 総フレーム数を設定（メソッドチェーン用）
   */
  withTotalFrames(totalFrames) {
    this.totalFrames = totalFrames;
    return this;
  }
  /**
   * ワーカーを設定（メソッドチェーン用）
   */
  withWorker(worker) {
    this.configuredWorker = worker;
    return this;
  }
  /**
   * ワーカースクリプトURLを設定（メソッドチェーン用）
   */
  withWorkerScript(url) {
    this.configuredWorkerScriptUrl = url;
    return this;
  }
  /**
   * AudioWorkletの使用を設定（メソッドチェーン用）
   */
  withAudioWorklet(useAudioWorklet = true) {
    this.configuredUseAudioWorklet = useAudioWorklet;
    return this;
  }
  /**
   * エンコーダーを開始（メソッドチェーン用のinitialize）
   */
  async start() {
    if (Object.keys(this.config).length === 0) {
      throw new WebCodecsEncoderError(
        "configuration-error" /* ConfigurationError */,
        "Configuration not set. Call configure() before start()."
      );
    }
    const options = {
      onProgress: this.onProgressCallback || void 0,
      onDetailedProgress: this.onDetailedProgressCallback || void 0,
      totalFrames: this.totalFrames,
      onError: this.onErrorCallback || void 0,
      onData: this.onDataCallback || void 0,
      worker: this.configuredWorker,
      workerScriptUrl: this.configuredWorkerScriptUrl,
      useAudioWorklet: this.configuredUseAudioWorklet
    };
    await this.initialize(options);
    return this;
  }
  /**
   * エンコーダーを終了（メソッドチェーン用のfinalize）
   */
  async finish() {
    return this.finalize();
  }
  /**
   * 詳細プログレス情報を計算
   */
  calculateDetailedProgress() {
    const currentTime = Date.now();
    const elapsedTimeMs = currentTime - this.processingStartTime;
    let processingFps = 0;
    if (this.lastProgressTime > 0) {
      const timeDiff = currentTime - this.lastProgressTime;
      const framesDiff = this.processedFramesInternal - this.lastProcessedFrames;
      if (timeDiff > 0) {
        processingFps = framesDiff / timeDiff * 1e3;
      }
    }
    if (processingFps > 0) {
      this.processingFpsHistory.push(processingFps);
      if (this.processingFpsHistory.length > this.maxFpsHistoryLength) {
        this.processingFpsHistory.shift();
      }
    }
    const averageProcessingFps = this.processingFpsHistory.length > 0 ? this.processingFpsHistory.reduce((a, b) => a + b, 0) / this.processingFpsHistory.length : 0;
    let estimatedRemainingMs;
    if (this.totalFrames && averageProcessingFps > 0) {
      const remainingFrames = this.totalFrames - this.processedFramesInternal;
      estimatedRemainingMs = remainingFrames / averageProcessingFps * 1e3;
    }
    this.lastProgressTime = currentTime;
    this.lastProcessedFrames = this.processedFramesInternal;
    return {
      processedFrames: this.processedFramesInternal,
      totalFrames: this.totalFrames,
      stage: this.currentStage,
      elapsedTimeMs,
      estimatedRemainingMs,
      processingFps,
      averageProcessingFps,
      droppedFrames: this.droppedFramesInternal,
      videoQueueSize: this.videoQueueSizeInternal,
      audioQueueSize: this.audioQueueSizeInternal,
      processedDataSize: this.processedDataSize
    };
  }
  async initialize(options) {
    this.checkStateForOperation("initialize", [
      "idle" /* Idle */,
      "error" /* Error */
    ]);
    this.setState("initializing" /* Initializing */);
    this.setStage("initializing" /* Initializing */);
    this.onErrorCallback = options?.onError || null;
    this.onProgressCallback = options?.onProgress || null;
    this.onDetailedProgressCallback = options?.onDetailedProgress || null;
    this.onDataCallback = options?.onData || null;
    this.totalFrames = options?.totalFrames;
    this.processingStartTime = Date.now();
    if (this.config.latencyMode === "realtime" && !this.onDataCallback) {
      const err = new WebCodecsEncoderError(
        "configuration-error" /* ConfigurationError */,
        "onData callback must be provided when latencyMode is 'realtime'."
      );
      throw err;
    }
    if (!_WebCodecsEncoder.isSupported()) {
      const err = new WebCodecsEncoderError(
        "not-supported" /* NotSupported */,
        "Required browser APIs (WebCodecs, Worker, etc.) are not supported."
      );
      this.handleError(err);
      throw err;
    }
    if (this.worker) {
      logger_default.warn(
        "WebCodecsEncoder already initialized or in progress. Call cancel() before re-initializing."
      );
      if (!this.isCancelled) {
        return Promise.resolve();
      }
    }
    this.isCancelled = false;
    this.processedFramesInternal = 0;
    this.submittedFramesInternal = 0;
    this.droppedFramesInternal = 0;
    this.videoQueueSizeInternal = 0;
    this.audioQueueSizeInternal = 0;
    this.nextVideoTimestamp = 0;
    this.nextAudioTimestamp = 0;
    return new Promise((resolve, reject) => {
      this.onInitialized = resolve;
      this.onInitializeError = reject;
      const start = async () => {
        try {
          if (options?.worker) {
            this.worker = options.worker;
          } else if (typeof process !== "undefined" && process.env?.NODE_ENV === "test") {
            this.worker = new Worker("data:application/javascript,", {
              type: "module"
            });
          } else {
            const script = await _WebCodecsEncoder.findWorkerScript(options?.workerScriptUrl);
            this.worker = new Worker(script, { type: "module" });
          }
          this.worker.onmessage = (event) => {
            this.handleWorkerMessage(event.data);
          };
          this.worker.onerror = (event) => {
            logger_default.error("MainThread: worker.onerror triggered. Event:", event);
            logger_default.error(
              "MainThread: worker.onerror event.message:",
              event.message
            );
            logger_default.error(
              "MainThread: worker.onerror event.filename:",
              event.filename
            );
            logger_default.error(
              "MainThread: worker.onerror event.lineno:",
              event.lineno
            );
            logger_default.error(
              "MainThread: worker.onerror event.colno:",
              event.colno
            );
            logger_default.error(
              "MainThread: worker.onerror event.error:",
              event.error
            );
            const err = new WebCodecsEncoderError(
              "worker-error" /* WorkerError */,
              `Worker error: ${event.message || "Unknown worker error"}`,
              event
            );
            this.handleError(err);
            this.onInitializeError?.(err);
            this.onFinalizedPromise?.reject(err);
            this.cleanupWorkerOnError();
          };
          if (options?.useAudioWorklet) {
            await this.setupAudioWorklet();
          }
          const initMessage = {
            type: "initialize",
            config: this.config,
            // Pass updated config
            totalFrames: this.totalFrames
          };
          this.worker.postMessage(initMessage);
        } catch (e) {
          const err = new WebCodecsEncoderError(
            "initialization-failed" /* InitializationFailed */,
            `Failed to initialize worker: ${e.message}`,
            e
          );
          this.handleError(err);
          this.onInitializeError?.(err);
          this.cleanupWorkerOnError();
        }
      };
      void start();
    });
  }
  static async findWorkerScript(customUrl) {
    if (customUrl) {
      return customUrl;
    }
    const isDev = _WebCodecsEncoder.isDevEnvironment();
    const publicPaths = [
      "/webcodecs-worker.js",
      // postinstall file (priority)
      "/worker.js"
      // common name
    ];
    for (const path of publicPaths) {
      const methods = isDev ? ["GET", "HEAD"] : ["HEAD"];
      for (const method of methods) {
        try {
          const response = await fetch(path, {
            method,
            // In dev, try with mode: 'no-cors' as fallback
            ...isDev && method === "GET" ? { mode: "no-cors" } : {}
          });
          if (response.ok || isDev && response.type === "opaque") {
            logger_default.log(
              `WebCodecsEncoder: Found worker at public path: ${path} (method: ${method})`
            );
            return path;
          }
        } catch (e) {
        }
      }
    }
    if (isDev) {
      const devPaths = [
        "/webcodecs-worker.js",
        "/public/webcodecs-worker.js"
        // Sometimes Vite exposes public files this way
      ];
      for (const path of devPaths) {
        try {
          const testWorker = new Worker(path, { type: "module" });
          testWorker.terminate();
          logger_default.log(
            `WebCodecsEncoder: Found worker in dev environment: ${path}`
          );
          return path;
        } catch (e) {
        }
      }
    }
    try {
      const packageWorkerUrl = new URL("./worker.js", import.meta.url);
      const response = await fetch(packageWorkerUrl, { method: "HEAD" });
      if (response.ok) {
        logger_default.log("WebCodecsEncoder: Using package worker file");
        return packageWorkerUrl;
      }
    } catch (e) {
    }
    logger_default.log(
      "WebCodecsEncoder: Creating inline worker (package worker not accessible)"
    );
    return _WebCodecsEncoder.createInlineWorker();
  }
  static isDevEnvironment() {
    return (
      // Node.js development
      typeof process !== "undefined" && process.env?.NODE_ENV === "development" || // Vite development (with type safety)
      typeof import.meta !== "undefined" && import.meta.env?.DEV === true || // Development server indicators
      typeof location !== "undefined" && (location.hostname === "localhost" || location.hostname === "127.0.0.1" || location.hostname.startsWith("192.168.") || location.port === "3000" || location.port === "5173" || // Vite default
      location.port === "4173")
    );
  }
  static createInlineWorker() {
    const setupInstructions = `
WebCodecs Encoder Worker Setup Required:

The worker file could not be loaded automatically. This is common in Vite/PWA projects.

Quick Setup:
1. Copy worker file to public directory:
   cp node_modules/webcodecs-encoder/dist/worker.js public/

2. Initialize with custom worker URL:
   const encoder = new WebCodecsEncoder(config);
   await encoder.initialize({
     workerScriptUrl: '/worker.js'
   });

For more setup options: https://github.com/romot-co/webcodecs-encoder#setup-for-vitepwa-projects
`;
    const inlineWorkerCode = `
// WebCodecs Encoder - Setup Required
self.postMessage({
  type: 'error',
  errorDetail: {
    type: 'WorkerError',
    message: ${JSON.stringify(setupInstructions)},
    stack: null
  }
});
`;
    const blob = new Blob([inlineWorkerCode], {
      type: "application/javascript"
    });
    return URL.createObjectURL(blob);
  }
  async findAudioWorkletScript() {
    const isDev = _WebCodecsEncoder.isDevEnvironment();
    const publicPaths = [
      "/webcodecs-audio-worklet.js",
      // postinstall file (priority)
      "/audio-worklet-processor.js"
      // alternative name
    ];
    for (const path of publicPaths) {
      const methods = isDev ? ["GET", "HEAD"] : ["HEAD"];
      for (const method of methods) {
        try {
          const response = await fetch(path, {
            method,
            ...isDev && method === "GET" ? { mode: "no-cors" } : {}
          });
          if (response.ok || isDev && response.type === "opaque") {
            logger_default.log(
              `WebCodecsEncoder: Found AudioWorklet processor at: ${path} (method: ${method})`
            );
            return path;
          }
        } catch (e) {
        }
      }
    }
    try {
      const packageUrl = new URL(
        "./audio-worklet-processor.js",
        import.meta.url
      );
      return packageUrl.href;
    } catch (e) {
      logger_default.warn(
        "WebCodecsEncoder: AudioWorklet processor not found, using fallback"
      );
      return "/webcodecs-audio-worklet.js";
    }
  }
  handleWorkerMessage(message) {
    if (this.isCancelled && message.type !== "cancelled" && message.type !== "error")
      return;
    switch (message.type) {
      case "initialized":
        this.setState("encoding" /* Encoding */);
        this.setStage("video-encoding" /* VideoEncoding */);
        this.actualVideoCodec = message.actualVideoCodec ?? null;
        this.actualAudioCodec = message.actualAudioCodec ?? null;
        this.onInitialized?.();
        this.onInitialized = null;
        this.onInitializeError = null;
        break;
      case "progress":
        this.processedFramesInternal = message.processedFrames + this.droppedFramesInternal;
        this.onProgressCallback?.(
          this.processedFramesInternal,
          message.totalFrames
        );
        if (this.onDetailedProgressCallback) {
          const detailedProgress = this.calculateDetailedProgress();
          this.onDetailedProgressCallback(detailedProgress);
        }
        break;
      case "detailedProgress":
        if (this.onDetailedProgressCallback) {
          this.onDetailedProgressCallback(message.progress);
        }
        break;
      case "queueSize":
        this.videoQueueSizeInternal = message.videoQueueSize;
        this.audioQueueSizeInternal = message.audioQueueSize;
        break;
      case "dataChunk":
        if (this.config.latencyMode === "realtime" && this.onDataCallback) {
          const { chunk, isHeader, container } = message;
          this.processedDataSize += chunk.byteLength;
          this.onDataCallback(chunk, void 0, isHeader, container);
        } else if (this.onDataCallback && this.config.latencyMode !== "realtime") {
        } else if (!this.onDataCallback && this.config.latencyMode === "realtime") {
          logger_default.warn(
            "WebCodecsEncoder: Received dataChunk in real-time mode, but no onData callback was provided."
          );
        }
        break;
      case "finalized":
        this.clearCancelTimeout();
        this.setState("disposed" /* Disposed */);
        if (this.config.latencyMode === "realtime") {
          this.onFinalizedPromise?.resolve(
            message.output === null ? null : new Uint8Array(0)
          );
        } else {
          if (message.output !== null) {
            this.processedDataSize = message.output.byteLength;
            this.onFinalizedPromise?.resolve(message.output);
          } else {
            const err2 = new WebCodecsEncoderError(
              "muxing-failed" /* MuxingFailed */,
              "Finalized with null output in non-realtime mode."
            );
            if (this.onFinalizedPromise) {
              this.onFinalizedPromise.reject(err2);
            } else {
              this.onInitializeError?.(err2);
            }
            this.handleError(err2);
          }
        }
        this.onFinalizedPromise = null;
        this.cleanupWorker();
        break;
      case "error":
        this.clearCancelTimeout();
        this.setState("error" /* Error */);
        const err = new WebCodecsEncoderError(
          message.errorDetail.type,
          message.errorDetail.message,
          message.errorDetail.stack
        );
        this.handleError(err);
        this.onInitializeError?.(err);
        this.onFinalizedPromise?.reject(err);
        this.cleanupWorkerOnError();
        break;
      case "cancelled":
        this.clearCancelTimeout();
        this.setState("disposed" /* Disposed */);
        logger_default.log("WebCodecsEncoder: Cancelled by worker.");
        const cancelErrWorker = new WebCodecsEncoderError(
          "cancelled" /* Cancelled */,
          "Operation cancelled by worker."
        );
        this.onInitializeError?.(cancelErrWorker);
        this.onFinalizedPromise?.reject(cancelErrWorker);
        this.cleanupWorker();
        break;
      default:
        const _exhaustiveCheck = message;
        logger_default.warn(
          "WebCodecsEncoder: Unknown message from worker:",
          _exhaustiveCheck
        );
    }
  }
  handleError(error) {
    console.error(
      `WebCodecsEncoder Error (${error.type}):`,
      error.message,
      error.cause || ""
    );
    this.onErrorCallback?.(error);
  }
  async addVideoFrame(frameSource, timestampOverride) {
    this.checkStateForOperation("add video frame", ["encoding" /* Encoding */]);
    if (!this.worker || this.isCancelled) {
      const err = new WebCodecsEncoderError(
        this.isCancelled ? "cancelled" /* Cancelled */ : "internal-error" /* InternalError */,
        this.isCancelled ? "Encoder cancelled" : "Encoder not initialized or already finalized"
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    try {
      const queueDepth = this.videoQueueSizeInternal;
      const maxDepth = this.config.maxQueueDepth ?? Infinity;
      if (this.config.dropFrames && queueDepth >= maxDepth) {
        frameSource.close();
        this.nextVideoTimestamp += 1e6 / this.config.frameRate;
        this.submittedFramesInternal++;
        this.droppedFramesInternal++;
        this.processedFramesInternal++;
        this.onProgressCallback?.(
          this.processedFramesInternal,
          this.totalFrames
        );
        return;
      }
      const timestamp = timestampOverride ?? this.nextVideoTimestamp;
      this.nextVideoTimestamp = timestamp + 1e6 / this.config.frameRate;
      this.submittedFramesInternal++;
      const message = {
        type: "addVideoFrame",
        frame: frameSource,
        // Send VideoFrame directly
        timestamp
      };
      this.worker.postMessage(message, [frameSource]);
    } catch (e) {
      const err = new WebCodecsEncoderError(
        "video-encoding-error" /* VideoEncodingError */,
        `Failed to post video frame: ${e.message}`,
        e
      );
      this.handleError(err);
      throw err;
    }
  }
  async addCanvasFrame(canvas) {
    const queueDepth = this.videoQueueSizeInternal;
    const maxDepth = this.config.maxQueueDepth ?? Infinity;
    if (this.config.dropFrames && queueDepth >= maxDepth) {
      this.nextVideoTimestamp += 1e6 / this.config.frameRate;
      this.submittedFramesInternal++;
      this.droppedFramesInternal++;
      this.processedFramesInternal++;
      this.onProgressCallback?.(this.processedFramesInternal, this.totalFrames);
      return;
    }
    const timestamp = this.nextVideoTimestamp;
    const frame = new VideoFrame(canvas, {
      timestamp,
      duration: 1e6 / this.config.frameRate
    });
    try {
      await this.addVideoFrame(frame);
    } catch (err) {
      frame.close();
      throw err;
    }
  }
  async addAudioBuffer(audioBuffer) {
    this.checkStateForOperation("add audio buffer", ["encoding" /* Encoding */]);
    if (!this.worker || this.isCancelled) {
      const err = new WebCodecsEncoderError(
        this.isCancelled ? "cancelled" /* Cancelled */ : "internal-error" /* InternalError */,
        this.isCancelled ? "Encoder cancelled" : "Encoder not initialized or already finalized"
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    if (!this.config.audioBitrate || this.config.audioBitrate <= 0 || this.config.channels <= 0 || this.config.sampleRate <= 0) {
      logger_default.warn(
        "Audio encoding is disabled (audioBitrate, channels or sampleRate is zero/negative). Skipping addAudioBuffer."
      );
      return Promise.resolve();
    }
    try {
      if (audioBuffer.numberOfChannels !== this.config.channels) {
        const err = new WebCodecsEncoderError(
          "configuration-error" /* ConfigurationError */,
          `AudioBuffer channel count (${audioBuffer.numberOfChannels}) does not match configured channels (${this.config.channels}).`
        );
        this.handleError(err);
        return Promise.reject(err);
      }
      const numChannels = audioBuffer.numberOfChannels;
      const planarData = [];
      const transferableBuffers = [];
      for (let i = 0; i < numChannels; i++) {
        const channelDataContent = audioBuffer.getChannelData(i);
        const channelDataCopy = new Float32Array(channelDataContent.length);
        channelDataCopy.set(channelDataContent);
        planarData.push(channelDataCopy);
        transferableBuffers.push(channelDataCopy.buffer);
      }
      const timestamp = this.nextAudioTimestamp;
      this.nextAudioTimestamp += audioBuffer.duration * 1e6;
      const message = {
        type: "addAudioData",
        audioData: planarData,
        timestamp,
        format: "f32-planar",
        // AudioBuffer.getChannelData は Float32Array を返すので planar f32
        sampleRate: audioBuffer.sampleRate,
        numberOfFrames: audioBuffer.length,
        // audioBuffer.length はフレーム数を返す
        numberOfChannels: numChannels
        // 使用するチャンネル数
      };
      this.worker.postMessage(message, transferableBuffers);
    } catch (e) {
      const err = new WebCodecsEncoderError(
        "audio-encoding-error" /* AudioEncodingError */,
        `Failed to add audio buffer: ${e.message}`,
        e
      );
      this.handleError(err);
      throw err;
    }
  }
  async addAudioData(audioData) {
    this.checkStateForOperation("add audio data", ["encoding" /* Encoding */]);
    if (!this.worker || this.isCancelled) {
      const err = new WebCodecsEncoderError(
        this.isCancelled ? "cancelled" /* Cancelled */ : "internal-error" /* InternalError */,
        this.isCancelled ? "Encoder cancelled" : "Encoder not initialized or already finalized"
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    if (!this.config.audioBitrate || this.config.audioBitrate <= 0 || this.config.channels <= 0 || this.config.sampleRate <= 0) {
      logger_default.warn(
        "Audio encoding is disabled (audioBitrate, channels or sampleRate is zero/negative). Skipping addAudioData."
      );
      return Promise.resolve();
    }
    try {
      if (audioData.numberOfChannels !== this.config.channels) {
        const err = new WebCodecsEncoderError(
          "configuration-error" /* ConfigurationError */,
          `AudioData channel count (${audioData.numberOfChannels}) does not match configured channels (${this.config.channels}).`
        );
        this.handleError(err);
        return Promise.reject(err);
      }
      const numChannels = audioData.numberOfChannels;
      const planarData = [];
      const transferableBuffers = [];
      for (let i = 0; i < numChannels; i++) {
        const plane = new Float32Array(audioData.numberOfFrames);
        await audioData.copyTo(plane, { planeIndex: i });
        planarData.push(plane);
        transferableBuffers.push(plane.buffer);
      }
      const timestamp = audioData.timestamp ?? this.nextAudioTimestamp;
      if (audioData.timestamp == null) {
        this.nextAudioTimestamp += audioData.numberOfFrames / this.config.sampleRate * 1e6;
      }
      const message = {
        type: "addAudioData",
        audioData: planarData,
        timestamp,
        format: "f32-planar",
        sampleRate: audioData.sampleRate,
        numberOfFrames: audioData.numberOfFrames,
        numberOfChannels: numChannels
      };
      this.worker.postMessage(message, transferableBuffers);
    } catch (e) {
      const err = new WebCodecsEncoderError(
        "audio-encoding-error" /* AudioEncodingError */,
        `Failed to add audio data: ${e.message}`,
        e
      );
      this.handleError(err);
      throw err;
    }
  }
  finalize() {
    this.checkStateForOperation("finalize", ["encoding" /* Encoding */]);
    this.setState("finalizing" /* Finalizing */);
    this.setStage("finalizing" /* Finalizing */);
    if (this.isCancelled) {
      const err = new WebCodecsEncoderError(
        "cancelled" /* Cancelled */,
        "Encoder cancelled"
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    if (!this.worker) {
      const err = new WebCodecsEncoderError(
        "internal-error" /* InternalError */,
        "Encoder not initialized or already finalized"
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    if (this.onFinalizedPromise) {
      logger_default.warn("Finalize already called.");
      const err = new WebCodecsEncoderError(
        "internal-error" /* InternalError */,
        "Finalize called multiple times."
      );
      this.handleError(err);
      return Promise.reject(err);
    }
    return new Promise((resolve, reject) => {
      this.onFinalizedPromise = { resolve, reject };
      const message = { type: "finalize" };
      this.worker.postMessage(message);
    });
  }
  cancel() {
    if (this.isCancelled || !this.worker) {
      logger_default.log("WebCodecsEncoder: Already cancelled or not initialized.");
      return;
    }
    this.isCancelled = true;
    logger_default.log("WebCodecsEncoder: Sending cancel signal to worker.");
    const message = { type: "cancel" };
    this.worker.postMessage(message);
    this.worker.onmessage = (event) => {
      if (event.data.type === "cancelled" || event.data.type === "error") {
        this.handleWorkerMessage(event.data);
      }
    };
    const cancelError = new WebCodecsEncoderError(
      "cancelled" /* Cancelled */,
      "Operation cancelled by user."
    );
    this.handleError(cancelError);
    this.onInitializeError?.(cancelError);
    this.onFinalizedPromise?.reject(cancelError);
    this.onInitializeError = null;
    this.onInitialized = null;
    this.onFinalizedPromise = null;
    this.clearCancelTimeout();
    this.cancelTimeoutId = setTimeout(() => {
      logger_default.warn(
        "WebCodecsEncoder: No 'cancelled' message received, terminating worker."
      );
      this.cleanupWorker();
    }, 5e3);
  }
  clearCancelTimeout() {
    if (this.cancelTimeoutId !== null) {
      clearTimeout(this.cancelTimeoutId);
      this.cancelTimeoutId = null;
    }
  }
  cleanupWorker() {
    this.clearCancelTimeout();
    if (this.worker) {
      this.worker.terminate();
      this.worker = null;
      logger_default.log("WebCodecsEncoder: Worker terminated and cleaned up.");
    }
    if (this.audioWorkletNode) {
      this.audioWorkletNode.port.postMessage({ close: true });
      this.audioWorkletNode.disconnect();
      this.audioWorkletNode = null;
    }
    if (this.audioContext) {
      this.audioContext.close();
      this.audioContext = null;
    }
    this.videoQueueSizeInternal = 0;
    this.audioQueueSizeInternal = 0;
    this.isCancelled = true;
    if (this.currentState !== "error" /* Error */) {
      this.setState("disposed" /* Disposed */);
    }
  }
  // Specific cleanup for errors to avoid double-terminating if worker itself errored.
  cleanupWorkerOnError() {
    this.cleanupWorker();
  }
  async setupAudioWorklet() {
    const AudioContextCtor = globalThis.AudioContext;
    if (!AudioContextCtor) {
      const err = new WebCodecsEncoderError(
        "not-supported" /* NotSupported */,
        "AudioContext not available for AudioWorklet."
      );
      this.handleError(err);
      throw err;
    }
    this.audioContext = new AudioContextCtor({
      sampleRate: this.config.sampleRate
    });
    const audioWorkletUrl = await this.findAudioWorkletScript();
    await this.audioContext.audioWorklet.addModule(audioWorkletUrl);
    this.audioWorkletNode = new AudioWorkletNode(
      this.audioContext,
      "encoder-audio-worklet",
      { numberOfInputs: 1, numberOfOutputs: 0 }
    );
    const { port1, port2 } = new MessageChannel();
    const connectMessage = {
      type: "connectAudioPort",
      port: port1
    };
    this.worker.postMessage(connectMessage, [port1]);
    this.audioWorkletNode.port.postMessage(
      { port: port2, sampleRate: this.config.sampleRate },
      [port2]
    );
  }
  getActualVideoCodec() {
    return this.actualVideoCodec;
  }
  getActualAudioCodec() {
    return this.actualAudioCodec;
  }
  getVideoQueueSize() {
    return this.videoQueueSizeInternal;
  }
  getAudioQueueSize() {
    return this.audioQueueSizeInternal;
  }
  getAudioWorkletNode() {
    return this.audioWorkletNode;
  }
};

// src/mediastream-recorder.ts
var MediaStreamRecorder = class {
  constructor(config) {
    this.config = config;
    this.recording = false;
    this.encoder = new WebCodecsEncoder(config);
  }
  static isSupported() {
    return typeof MediaStreamTrackProcessor !== "undefined" && WebCodecsEncoder.isSupported();
  }
  async startRecording(stream, options) {
    if (this.recording) {
      throw new Error("MediaStreamRecorder: already recording.");
    }
    this.onErrorCallback = options?.onError;
    const initializeOptions = {
      ...this.config.firstTimestampBehavior && {
        firstTimestampBehavior: this.config.firstTimestampBehavior
      },
      ...options ?? {}
    };
    await this.encoder.initialize(initializeOptions);
    this.recording = true;
    const [vTrack] = stream.getVideoTracks();
    const [aTrack] = stream.getAudioTracks();
    if (vTrack) {
      this.videoTrack = vTrack;
      const processor = new MediaStreamTrackProcessor({
        track: vTrack
      });
      this.videoReader = processor.readable.getReader();
      this.processVideo();
    }
    if (aTrack) {
      this.audioTrack = aTrack;
      if (options?.useAudioWorklet) {
        const node = this.encoder.getAudioWorkletNode();
        if (!node) {
          throw new Error(
            "MediaStreamRecorder: AudioWorkletNode not available from encoder."
          );
        }
        const ctx = node.context;
        this.audioSource = ctx.createMediaStreamSource(stream);
        this.audioSource.connect(node);
      } else {
        const processor = new MediaStreamTrackProcessor({
          track: aTrack
        });
        this.audioReader = processor.readable.getReader();
        this.processAudio();
      }
    }
  }
  async processVideo() {
    if (!this.videoReader) return;
    const reader = this.videoReader;
    try {
      while (this.recording) {
        const { value, done } = await reader.read();
        if (done || !value) {
          if (this.recording) {
            await this.stopRecording();
          }
          break;
        }
        try {
          await this.encoder.addVideoFrame(value);
        } finally {
          value.close();
        }
      }
    } catch (err) {
      this.cancel();
      if (this.onErrorCallback) {
        this.onErrorCallback(err);
      } else {
        throw err;
      }
    } finally {
      reader.cancel();
      this.videoReader = void 0;
    }
  }
  async processAudio() {
    if (!this.audioReader) return;
    const reader = this.audioReader;
    try {
      while (this.recording) {
        const { value, done } = await reader.read();
        if (done || !value) {
          if (this.recording) {
            await this.stopRecording();
          }
          break;
        }
        try {
          await this.encoder.addAudioData(value);
        } finally {
          value.close();
        }
      }
    } catch (err) {
      this.cancel();
      if (this.onErrorCallback) {
        this.onErrorCallback(err);
      } else {
        throw err;
      }
    } finally {
      reader.cancel();
      this.audioReader = void 0;
    }
  }
  async stopRecording() {
    if (!this.recording) {
      throw new Error("MediaStreamRecorder: not recording.");
    }
    this.recording = false;
    this.videoReader?.cancel();
    this.audioReader?.cancel();
    this.audioSource?.disconnect();
    this.videoTrack?.stop();
    this.audioTrack?.stop();
    this.videoReader = void 0;
    this.audioReader = void 0;
    this.audioSource = void 0;
    this.videoTrack = void 0;
    this.audioTrack = void 0;
    return await this.encoder.finalize();
  }
  cancel() {
    if (!this.recording) return;
    this.recording = false;
    this.videoReader?.cancel();
    this.audioReader?.cancel();
    this.audioSource?.disconnect();
    this.videoTrack?.stop();
    this.audioTrack?.stop();
    this.videoReader = void 0;
    this.audioReader = void 0;
    this.audioSource = void 0;
    this.videoTrack = void 0;
    this.audioTrack = void 0;
    this.encoder.cancel();
  }
  getActualVideoCodec() {
    return this.encoder.getActualVideoCodec();
  }
  getActualAudioCodec() {
    return this.encoder.getActualAudioCodec();
  }
};
export {
  MediaStreamRecorder,
  WebCodecsEncoder
};
//# sourceMappingURL=index.js.map